package info.textgrid.middleware;

import java.io.IOException;
import java.io.StringWriter;
import java.math.BigInteger;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TimeZone;
import java.util.logging.Logger;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import org.elasticsearch.action.get.GetRequest;
import org.elasticsearch.action.get.GetResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.search.fetch.subphase.FetchSourceContext;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import info.textgrid.clients.AuthClient;
import info.textgrid.clients.tgauth.AuthClientException;
import info.textgrid.middleware.oaipmh.GetRecordType;
import info.textgrid.middleware.oaipmh.HeaderType;
import info.textgrid.middleware.oaipmh.ListMetadataFormatsType;
import info.textgrid.middleware.oaipmh.MetadataFormatType;
import info.textgrid.middleware.oaipmh.MetadataType;
import info.textgrid.middleware.oaipmh.RecordType;
import info.textgrid.middleware.oaipmh.Resource;
import info.textgrid.middleware.oaipmh.ResumptionTokenType;
import info.textgrid.namespaces.middleware.tgauth.ProjectInfo;

/**
 * <p>
 * OAIPMHUtillities is responsible to implement all necessary functions to implement the arguments
 * of a request.
 * </p>
 * 
 * @author Maximilian Brodhun, SUB Göttingen
 * @author Stefan E. Funk, SUB Göttingen
 * @version 2022-07-06
 * @since 2014-01-29
 */
public class OAIPMHUtilities {

  private static Logger log = Logger.getLogger(OAIPMHUtilities.class.getName());

  // **
  // FINALS
  // **

  public static final String OAIPMH_NAMESPACE = "http://www.openarchives.org/OAI/2.0/";
  public static final String OAIPMH_SCHEMA_LOCATION =
      "http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd";

  public static final String OAIDC_PREFIX = "oai_dc";
  public static final String OAIDC_NAMESPACE = "http://www.openarchives.org/OAI/2.0/oai_dc/";
  public static final String OAIDC_SCHEMA_LOCATION =
      "http://www.openarchives.org/OAI/2.0/oai_dc.xsd";

  public static final String DC_NAMESPACE = "http://purl.org/dc/elements/1.1/";
  public static final String DC_SCHEMA_LOCATION =
      "https://dublincore.org/schemas/xmls/simpledc20021212.xsd";

  public static final String METS_NAMESPACE = "http://www.loc.gov/METS/";
  public static final String METS_SCHEMA_LOCATION = "http://www.loc.gov/standards/mets/mets.xsd";

  public static final String OPEN_AIRE_NAMESPACE = "http://namespace.openaire.eu/oaf";
  public static final String OPEN_AIRE_SCHEMA_LOCATION =
      "https://www.openaire.eu/schema/1.0/oaf-1.0.xsd";

  public static final String IDIOM_IMAGE_NAMESPACE = "http://www.loc.gov/METS/";
  public static final String IDIOM_IMAGE_SCHEMA_LOCATION =
      "http://www.loc.gov/standards/mets/mets.xsd";

  public static final String DATACITE_PREFIX = "datacite";
  public static final String DATACITE_NAMESPACE = "http://datacite.org/schema/kernel-3";
  public static final String DATACITE_SCHEMA_LOCATION =
      "https://schema.datacite.org/meta/kernel-3/metadata.xsd";

  public static final String ES_DIVIDER_CHAR = ".";
  public static final String ES_DIVIDER_REGEXP = "\\.";

  public static final String PREFIX_DEVIDER = ":";
  public static final String PREFIX_HDL = "hdl";
  public static final String PREFIX_TEXTGRID = "textgrid";
  public static final Set<String> IDENTIFIER_PREFIXES_TO_FILTER = new HashSet<String>(Arrays.asList(
      PREFIX_HDL,
      PREFIX_TEXTGRID));

  // **
  // CLASS
  // **

  private String rangeField;
  private long resultSize;
  private boolean foundItems;
  private String formatField;
  private ResumptionTokenType resTokenForResponse;
  private String formatToFilter;
  private int searchResponseSize;

  // **
  // STATICS
  // **

  /**
   * @param verb
   * @return
   */
  public static boolean isOAIPMH_RequestArgument(String verb) {
    String allowedArgument = "^[identifier,from,until,set,resumptionToken,metadataPrefix]*$";
    return verb.matches(allowedArgument);
  }

  /**
   * <p>
   * Since TextGrid just supports DC the only returned MetadataFormat is DublinCore.
   * </p>
   * 
   * @return
   */
  public MetadataFormatType setMetadataFormats() {

    MetadataFormatType tgDublinCore = new MetadataFormatType();
    tgDublinCore.setMetadataNamespace("http://www.openarchives.org/OAI/2.0/oai_dc/");
    tgDublinCore.setMetadataPrefix("oai_dc");
    tgDublinCore.setSchema("http://www.openarchives.org/OAI/2.0/oai_dc.xsd");

    return tgDublinCore;
  }

  /**
   * @return
   * @throws AuthClientException
   */
  public static List<ProjectInfo> getProjectList() throws AuthClientException {

    AuthClient auth = new AuthClient();
    List<ProjectInfo> projectInfos = auth.getAllProjects();
    auth.getAllProjects();

    return projectInfos;
  }

  /**
   * @param projectID
   * @return
   * @throws AuthClientException
   */
  public static String getProjectName(String projectID) throws AuthClientException {
    AuthClient auth = new AuthClient();
    return auth.getProjectInfo(projectID).getName();
  }

  /**
   * <p>
   * Producing the list of all metadata formats.
   * </p>
   * 
   * @return tgRepMetadaFormats
   */
  public ListMetadataFormatsType setMetadataFormatList() {

    ListMetadataFormatsType tgRepMetadataFormats = new ListMetadataFormatsType();
    MetadataFormatType tgDublinCore = setMetadataFormats();
    tgRepMetadataFormats.getMetadataFormat().add(tgDublinCore);

    return tgRepMetadataFormats;
  }

  /**
   * <p>
   * Calculating the current date and give it back as XMLGregorianCalendar.
   * </p>
   * 
   * @return now
   * @throws DatatypeConfigurationException
   */
  public static XMLGregorianCalendar getXMLGregorianCalendarNow()
      throws DatatypeConfigurationException {

    GregorianCalendar gregorianCalendar = new GregorianCalendar();
    DatatypeFactory datatypeFactory = DatatypeFactory.newInstance();
    XMLGregorianCalendar now = datatypeFactory.newXMLGregorianCalendar(gregorianCalendar);

    return now;
  }

  /**
   * <p>
   * Converting a given string representing a date value into the date format required for OAIPMH
   * and give it back as XMLGregorianCalendar.
   * </p>
   * 
   * @param originalDateTimeString
   * @return xmlCal
   * @throws ParseException
   * @throws DatatypeConfigurationException
   */
  public static XMLGregorianCalendar convertDateFormat(String originalDateTimeString)
      throws ParseException, DatatypeConfigurationException {

    log.fine("original date time string: " + originalDateTimeString);

    SimpleDateFormat tgItemTime = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.S");
    Date date = tgItemTime.parse(originalDateTimeString);
    SimpleDateFormat outFormatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
    outFormatter.setTimeZone(TimeZone.getTimeZone("UTC"));
    String output2 = outFormatter.format(date);
    XMLGregorianCalendar xmlCal = DatatypeFactory.newInstance().newXMLGregorianCalendar(output2);

    return xmlCal;
  }

  /**
   * <p>
   * Converting a given string representing a date value into the date format required for OAIPMH
   * and give it back as string.
   * </p>
   * 
   * @param originalDateTimeString
   * @return dateOutputAsString
   * @throws ParseException
   */
  public static String datestampAsString(String originalDateTimeString) throws ParseException {

    SimpleDateFormat tgItemTime = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.S");
    Date date = tgItemTime.parse(originalDateTimeString);
    SimpleDateFormat outFormatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
    outFormatter.setTimeZone(TimeZone.getTimeZone("UTC"));
    String dateOutputAsString = outFormatter.format(date);

    return dateOutputAsString;
  }

  /**
   * <p>
   * Returns a resumptionToken for a search, if necessary.
   * </p>
   * 
   * TODO Check if we really must put all these references in!
   * 
   * @param completeListSize
   * @param resumptionToken
   * @param cursorCollector
   * @param scrollID
   * @param searchResponseSize
   * @param i
   * @return A resumption token if applicable, null otherwise.
   */
  public static ResumptionTokenType getResumptionToken(final long completeListSize,
      final String resumptionToken, Map<String, Integer> cursorCollector, final String scrollID,
      final int searchResponseSize, final int i) {

    log.fine("Creating a ResumptionToken:\n " + "CompleteListSize: " + completeListSize + "\n"
        + "ResumptionToken: " + resumptionToken + "\n" + "CursorCollector: " + cursorCollector
        + "\n" + "ScrollID: " + scrollID + "\n" + "SearchResponseSize: " + searchResponseSize + "\n"
        + "IterationStep: " + i);

    int cursor;

    // Three cases here:
    // 1. Complete list size is > searchResponseSize and a token already is existing: we need to
    // check hash map!
    // 2. Complete list size is > searchResponseSize and we have no token: we do need one!
    // 3. Complete list size is <= searchResponseSize (we do not need a token! do nothing!)
    if (completeListSize > searchResponseSize) {
      ResumptionTokenType resTokenForResponse = new ResumptionTokenType();

      if (resumptionToken != null && cursorCollector.containsKey(resumptionToken)) {

        cursor = cursorCollector.get(resumptionToken).intValue() + i;
        if (cursor == searchResponseSize) {
          // TODO Why not "cursor += searchResponseSize;" ??
          cursor = 2 * searchResponseSize;
        }
        resTokenForResponse.setCursor(BigInteger.valueOf((long) cursor));
        cursorCollector.put(scrollID, cursor);
      } else {
        resTokenForResponse.setCursor(BigInteger.valueOf(searchResponseSize));
        cursorCollector.put(scrollID, searchResponseSize);
        cursor = searchResponseSize;
        resTokenForResponse.setCursor(BigInteger.valueOf((long) cursor));

        log.fine("New ResumptionToken created from scroll ID: " + cursorCollector.get(scrollID));
      }

      // Set resumption token string if cursor is less then complete list size: More objects can be
      // delivered!
      if (cursor < completeListSize) {
        resTokenForResponse.setValue(scrollID);
      }

      // Remove resumption token string if cursor is bigger then or equal complete list size: This
      // is the last response, no more objects are available!
      if (cursor >= completeListSize) {
        resTokenForResponse.setValue("");
      }

      // Set complete list size.
      resTokenForResponse.setCompleteListSize(BigInteger.valueOf(completeListSize));

      return resTokenForResponse;
    }

    return null;
  }

  /**
   * @param dateToValidate
   * @return
   */
  public static boolean isThisDateValid(String dateToValidate) {

    SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.S");

    dateFormat.setLenient(false);
    try {
      dateFormat.parse(dateToValidate.trim());
    } catch (ParseException pe) {
      return false;
    }
    return true;
  }

  /**
   * @param dateToValidate
   * @return
   */
  public static boolean isThisDateValidToOtherTimeStamp(String dateToValidate) {

    SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");

    dateFormat.setLenient(false);
    dateFormat.setLenient(false);
    try {
      dateFormat.parse(dateToValidate.trim());
    } catch (ParseException pe) {
      return false;
    }
    return true;
  }

  /**
   * 
   */
  private static final String[] formats = {
      "yyyy-MM-dd'T'HH:mm:ss'Z'", "yyyy-MM-dd'T'HH:mm:ssZ",
      "yyyy-MM-dd'T'HH:mm:ss", "yyyy-MM-dd'T'HH:mm:ss.SSS'Z'",
      "yyyy-MM-dd'T'HH:mm:ss.SSSZ", "yyyy-MM-dd HH:mm:ss",
      "MM/dd/yyyy HH:mm:ss", "MM/dd/yyyy'T'HH:mm:ss.SSS'Z'",
      "MM/dd/yyyy'T'HH:mm:ss.SSSZ", "MM/dd/yyyy'T'HH:mm:ss.SSS",
      "MM/dd/yyyy'T'HH:mm:ssZ", "MM/dd/yyyy'T'HH:mm:ss",
      "yyyy:MM:dd HH:mm:ss", "yyyyMMdd",};

  /**
   * @param d
   * @return
   */
  public static String getFormatOfDate(String d) {

    if (d != null) {
      for (String parse : formats) {
        SimpleDateFormat sdf = new SimpleDateFormat(parse);
        try {
          sdf.parse(d);
          return parse;
        } catch (ParseException e) {
          // FIXME Handle Exception!
        }
      }
    }

    return d;
  }

  /**
   * <p>
   * Loads the field with MORE params from an array!
   * </p>
   * 
   * @param resultFromGetRequestInES
   * @param fields
   * @return Returns MORE params as a list!
   */
  public static List<String> fieldLoader(JSONObject resultFromGetRequestInES, String[] fields) {

    log.fine("JSON: " + (resultFromGetRequestInES == null ? "NULL" : resultFromGetRequestInES));
    log.fine("FIELDS: " + fields);

    List<String> fieldResults = new ArrayList<String>();

    int count = 0;
    for (String field : fields) {
      String[] fieldPathForESIndex = field.split(ES_DIVIDER_REGEXP);

      log.fine("field[" + count++ + "]: " + field);

      JSONObject singlePath = resultFromGetRequestInES;
      try {
        for (int i = 0; i < fieldPathForESIndex.length; i++) {
          if (i < fieldPathForESIndex.length - 1) {
            singlePath = singlePath.getJSONObject(fieldPathForESIndex[i]);
          } else if (fieldPathForESIndex.length == 1) {
            JSONObject resultiDingsda =
                resultFromGetRequestInES.getJSONObject(fieldPathForESIndex[i]);
            fieldResults.add(resultiDingsda.toString());
          } else {
            String res = singlePath.get(fieldPathForESIndex[i]).toString();
            if (res.startsWith("[")) {
              JSONArray array = new JSONArray(res);
              for (int j = 0; j < array.length(); j++) {
                String find = array.getString(j).toString();
                fieldResults.add(find);
              }
            } else {
              fieldResults.add(singlePath.get(fieldPathForESIndex[i]).toString());
            }
          }
        }
      } catch (JSONException e) {
        log.fine("IGNORING JSON ERROR: " + e.getMessage());
      }
    }

    log.fine("field results: " + fieldResults);

    return fieldResults;
  }

  /**
   * <p>
   * Loads the field with ONE param!
   * </p>
   * 
   * @param resultFromGetRequestInES
   * @param field
   * @return Returns only ONE String!
   */
  public static String fieldLoader(JSONObject resultFromGetRequestInES, String field) {

    String fieldResults = "";

    String[] fieldPathForESIndex = field.split(ES_DIVIDER_REGEXP);
    JSONObject singlePath = resultFromGetRequestInES;
    try {
      for (int i = 0; i < fieldPathForESIndex.length; i++) {
        if (i < fieldPathForESIndex.length - 1) {
          singlePath = singlePath.getJSONObject(fieldPathForESIndex[i]);
        } else if (fieldPathForESIndex.length == 1) {
          fieldResults = resultFromGetRequestInES.get(fieldPathForESIndex[i]).toString();
        } else {
          fieldResults = singlePath.get(fieldPathForESIndex[i]).toString();
        }
      }
    } catch (JSONException e) {
      log.fine("IGNORING JSON ERROR: " + e.getMessage());
    }

    return fieldResults;
  }

  /**
   * @param theElasticSearchIndexToUse
   * @param idInElasticSearchIndex The ID to get the ES record from.
   * @param includes Including the fields to fetch from the elastic search index.
   * @param excludes Excluding the fields to fetch from the elastic search index.
   * @return
   * @throws IOException
   */
  protected static GetResponse getRcordByIDFromElasticSearch(String theElasticSearchIndexToUse,
      String idInElasticSearchIndex, String[] includes, String[] excludes) throws IOException {

    // Setting the source context for fetching the fields from the elastic search index
    FetchSourceContext fetchSourceContext = new FetchSourceContext(true, includes, excludes);

    // Building the getRequest against the elastic search index
    GetRequest getRequest =
        new GetRequest(theElasticSearchIndexToUse, OAI_ESClient.getEsType(), idInElasticSearchIndex)
            .fetchSourceContext(fetchSourceContext);

    // Declaration of the result from the get-request
    GetResponse esResultObject = OAI_ESClient.getEsClient().get(getRequest, RequestOptions.DEFAULT);

    return esResultObject;
  }

  /**
   * @param resource
   * @throws JAXBException
   */
  public static void marshal(Resource resource) throws JAXBException {
    JAXBContext jc = JAXBContext.newInstance(Resource.class);
    Marshaller marshaller = jc.createMarshaller();
    StringWriter stringWriter = new StringWriter();
    marshaller.marshal(resource, stringWriter);
    marshaller.marshal(resource, System.out);
  }

  /**
   * @param metadata
   * @param header
   * @return
   */
  public static GetRecordType getRecordType(Object metadata, HeaderType header) {

    MetadataType metadataType = new MetadataType();
    RecordType recordType = new RecordType();
    GetRecordType getRecordType = new GetRecordType();

    metadataType.setAny(metadata);
    recordType.setMetadata(metadataType);
    recordType.setHeader(header);
    getRecordType.setRecord(recordType);

    return getRecordType;
  }

  /**
   * @param propertyName
   * @param propFile
   * @return
   */
  @Deprecated
  public static String[][] fetchArrayFromPropFile(String propertyName, Properties propFile) {

    // FIXME Why get a 2D array and only use the first element in RecordDelivererDatacite?
    // BTW Where does the ";" come from anyway?

    // Get array split up by the semicolon.
    String[] a = propFile.getProperty(propertyName).split(";");

    // Create the two dimensional array with correct size.
    String[][] array = new String[a.length][a.length];

    // Combine the arrays split by semicolon and comma.
    for (int i = 0; i < a.length; i++) {
      array[i] = a[i].split(",");
    }

    return array;
  }

  /**
   * @param propertyName
   * @param propFile
   * @return
   */
  public static String[] fetchFieldsFromPropFile(String propertyName, Properties propFile) {
    // Trim with whitespaces, so we get the array with results trimmed.
    return propFile.getProperty(propertyName).split("\\s*,\\s*");
  }

  /**
   * <p>
   * Check identifier for prefixes: Remove "textgrid:" and "hdl:" prefixes.
   * </p>
   * 
   * @param identifier
   * @return The identifier without prefix, if contained in defined prefix list.
   */
  public static String omitPrefixFromIdentifier(final String identifier) {

    String result = identifier;

    if (identifier != null && !identifier.isEmpty() && identifier.contains(PREFIX_DEVIDER)) {
      String prefix[] = identifier.split(PREFIX_DEVIDER);
      String px = prefix[0];
      String id = prefix[1];
      if (IDENTIFIER_PREFIXES_TO_FILTER.contains(px)) {
        result = id;
        log.fine("incoming id has been filtered to: " + result);
      }
    }

    log.fine("  ##  omit prefix? " + identifier + " --> " + result);

    return result;
  }

  /**
   * <p>
   * Create header type of OAI-PMH response.
   * </p>
   * 
   * @param dateOfCreation The date of the resource's creation.
   * @param identifier The identifier to set in the header. Please see to it, that it is correct
   *        here, and have not to be changed in any way.
   * @param setSpec The ID of the set spec.
   * @return
   */
  public static HeaderType computeResponseHeader(final String dateOfCreation,
      final String identifier, final String setSpec) {

    HeaderType result = new HeaderType();

    result.setIdentifier(identifier);
    result.setDatestamp(dateOfCreation);

    log.info(
        "Header (DoC / ID / setSpec): " + dateOfCreation + " / " + identifier + " / " + setSpec);

    if (setSpec != null && !setSpec.isEmpty()) {
      result.getSetSpec().add(setSpec);
    }

    return result;
  }

  /**
   * <p>
   * Get the setSpec, if given, check for prefix, if not given, use prefix and given ID.
   * </p>
   * 
   * @param theSetSpec
   * @param theSpecPrefix
   * @param theID
   * @return A valid setSpec value for the set.
   */
  protected static String getSetSpec(String theSetSpec, String theSpecPrefix, String theID) {

    String result = theSetSpec;

    // Check given setSpec value, add ID if empty.
    if (result == null || result.isEmpty() || result.equals(theSpecPrefix)) {
      result = theID;
    }

    // Then check for prefix, add prefix, if necessary.
    if (!result.startsWith(theSpecPrefix)) {
      result = theSpecPrefix + result;
    }

    return result;
  }

  // **
  // GETTERS & SETTERS
  // **

  /**
   * @return
   */
  public String getRangeField() {
    return this.rangeField;
  }

  /**
   * @param rangeField
   */
  public void setRangeField(String rangeField) {
    this.rangeField = rangeField;
  }

  /**
   * @return
   */
  public long getResultSize() {
    return this.resultSize;
  }

  /**
   * @param resultSize
   */
  public void setResultSize(long resultSize) {
    this.resultSize = resultSize;
  }

  /**
   * @return
   */
  public boolean isFoundItems() {
    return this.foundItems;
  }

  /**
   * @param foundItems
   */
  public void setFoundItems(boolean foundItems) {
    this.foundItems = foundItems;
  }

  /**
   * @return
   */
  public ResumptionTokenType getResTokenForResponse() {
    return this.resTokenForResponse;
  }

  /**
   * @param resTokenForResponse
   */
  public void setResTokenForResponse(ResumptionTokenType resTokenForResponse) {
    this.resTokenForResponse = resTokenForResponse;
  }

  /**
   * @return
   */
  public String setFormatField() {
    return this.formatField;
  }

  /**
   * @param formatField
   */
  public void setFormatField(String formatField) {
    this.formatField = formatField;
  }

  /**
   * @return
   */
  public String getFormatToFilter() {
    return this.formatToFilter;
  }

  /**
   * @param formatToFilter
   */
  public void setFormatToFilter(String formatToFilter) {
    this.formatToFilter = formatToFilter;
  }

  /**
   * @return
   */
  public int getSearchResponseSize() {
    return this.searchResponseSize;
  }

  /**
   * @param searchResponseSize
   */
  public void setSearchResponseSize(int searchResponseSize) {
    this.searchResponseSize = searchResponseSize;
  }

}
